INPUT_IS_FILE_PATH_LIST = "__INPUT_IS_FILE_PATH_LIST__"
OUTPUT_IS_EXCEL_FILE = "__OUTPUT_IS_EXCEL_FILE__"
SKIP_CONVERT_TO_JSON = "__SKIP_CONVERT_TO_JSON__"

GRCH37 = "GRCh37"
GRCH38 = "GRCh38"
BUILDS = [GRCH37, GRCH38]
